# coding: utf-8
u"""Вспомогательные средства для работы с компонентами M3."""
import inspect
import sys

from django.core.exceptions import NON_FIELD_ERRORS
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db.models.base import Model
from objectpack.exceptions import ValidationError as ObjectPackValidationError
from objectpack.models import ModelProxy


def convert_validation_error_to(exc, new_line=u'<br/>'):
    """Декоратор, преобразующий исключение
    *django.core.exceptions.ValidationError*, генерируемое в декорируемой
    функции, в исключение, указанное в аргументе exc путем объединения всех
    сообщений об ошибках из *ValidationError.message_dict* в одно сообщение, по
    одной ошибке на строку.

    .. note::

        Пример использования см. в ValidationMixin.

    :param exc: класс исключения, к которому будет преобразовываться
        ValidationError
    :type exc: subclass of Exception
    :param unicode new_line: разделитель строк в сообщении об ошибке
    """
    def get_model_instance(error):
        # Достанем из стека вызовов объект модели, в которой было
        # вызвано исключение. Из него будем брать verbose_name полей.
        tb = sys.exc_info()[-1]  # traceback
        # Фрейм, в котором сгенерировано исключение, будет последним.
        error_frame = inspect.getinnerframes(tb)[-1][0]
        # f_locals - локальные переменные функции, в т.ч. аргументы.
        if 'self' not in error_frame.f_locals:
            raise
        model_instance = error_frame.f_locals['self']
        return model_instance

    def get_messages_from_dict(model_instance, data):
        result = []
        for field_name, field_errors in data.iteritems():
            if field_name == NON_FIELD_ERRORS:
                result.append(
                    new_line.join(
                        u'- {0}'.format(err) for err in field_errors
                    )
                )
            else:
                model_meta = model_instance._meta
                model_field = model_meta.get_field_by_name(
                    field_name
                )[0]
                verbose_name = model_field.verbose_name
                result.append(new_line.join(
                    u'- {0}: {1}'.format(verbose_name, err)
                    for err in field_errors
                ))
        return result

    def get_messages_from_list(messages):
        result = [
            u'- ' + message
            for message in messages
        ]
        return result

    assert issubclass(exc, Exception), type(exc)
    new_line = unicode(new_line)

    def decorator(func):
        assert inspect.ismethod(func) or inspect.isfunction(func)

        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except DjangoValidationError, e:
                model_instance = get_model_instance(e)
                if not isinstance(model_instance, Model):
                    raise

                if hasattr(e, 'message_dict'):
                    title = u'На форме имеются некорректно заполненные поля:'
                    messages = [title] + get_messages_from_dict(model_instance,
                                                                e.message_dict)
                else:
                    title = u'При проверке данных найдены ошибки:'
                    messages = [title] + get_messages_from_list(e.messages)
                messages.insert(1, u'')

                raise exc(new_line.join(messages))

        return wrapper

    return decorator


class ModelProxyValidationMixin(object):

    u"""Примесь для составных прокси-моделей objectpack'а.

    Добавляет к *objectpack.models.ModelProxy* возможность валидации данных
    перед сохранением. Валидация осуществляется, как и в моделях Django, путем
    запуска метода *full_clean()* для каждой зависимой модели, порядок запуска
    соответствует порядку указания моделей в *relations*. После выполнения
    валидации зависимых моделей запускается *full_clean()* основной модели.
    """

    def full_clean(self, exclude=None):
        u"""Валидация данных составной модели.

        :param list exclude: Список полей составной модели, исключаемых из
            проверки. Например: 'relation', 'relation1.relation2.name'.

            .. note::
               При многоуровневой вложенности вложенная составная модель также
               должна поддерживать валидацию данных (иметь метод
               *full_clean()*).

        :raises django.core.exceptions.ValidationError: Если данные составной
            модели некорректны.
        """
        exclude = exclude or []
        errors = {}

        models = (
            name
            for name in (list(self.relations) + [self.model.__name__.lower()])
            if not exclude or name not in exclude
        )

        for model_name in models:
            model = reduce(getattr, model_name.split('.'), self)

            # Формирование списка исключамых из проверки полей
            if model_name in exclude:
                # модель исключена полностью
                continue
            model_exclude = []
            for e in exclude:
                e = e.split('.', 1)
                if len(e) != 2:
                    continue
                m, f = e
                if m != model_name:
                    continue
                if not hasattr(self, m):
                    raise ValueError('Invalid exclude {}.{}'.format(m, f))
                else:
                    model_exclude.append(f)

            try:
                model.full_clean(model_exclude)
            except DjangoValidationError, error:
                model_errors = {}
                error.update_error_dict(model_errors)
                errors.update({
                    '.'.join((model_name, field)): messages
                    for field, messages in model_errors.iteritems()
                })

        if errors:
            raise DjangoValidationError(errors)


class BaseModelProxy(ModelProxyValidationMixin, ModelProxy):

    u"""Базовый класс для составных прокси моделей с валидацией.

    .. seealso::
        - :py:class: objectpack.models.ModelProxy
        - :py:class: ModelProxyValidationMixin
    """


class PackValidationMixin(object):

    u"""Примесь к пакам из objectpack, добавляющая валидацию моделей.

    Перед сохранением объекта в методе *save_row()* пака выполняется проверка
    данных путем вызова метода *full_clean()* сохраняемого объекта.

    .. note::
       При использовании в паке составной модели
       (*objectpack.models.ModelProxy*) в такой модели должен быть реализован
       метод *full_clean()* (см. *ModelProxyValidationMixin* и
       *BaseModelProxy*).

    Пример использования:

       class UnitPack(PackValidationMixin, TreeObjectPack):
           ...

       class PeriodPack(PackValidationMxin, ObjectPack):
           ....
    """

    @convert_validation_error_to(ObjectPackValidationError)
    def save_row(self, obj, create_new, request, context):
        u"""Вызывает проверку данных перед их сохранением в БД."""
        obj.full_clean()

        super(PackValidationMixin, self).save_row(
            obj, create_new, request, context
        )
